@@ -1,15 +1,33 @@
+from __future__ import division
+
 import os
 import subprocess
 from tempfile import TemporaryFile, NamedTemporaryFile
-from StringIO import StringIO
 import wave
 import audioop
+import sys
 
-from .utils import _fd_or_path_or_tempfile, db_to_float
-from .exceptions import TooManyMissingFrames
-from .exceptions import InvalidDuration
-from pydub.utils import ratio_to_db
+try:
+    from StringIO import StringIO
+except:
+    from io import StringIO, BytesIO
 
+from .utils import (
+    _fd_or_path_or_tempfile,
+    db_to_float,
+    ratio_to_db,
+)
+from .exceptions import (
+    TooManyMissingFrames,
+    InvalidDuration,
+    InvalidID3TagVersion,
+    InvalidTag,
+)
+
+if sys.version_info >= (3, 0):
+    basestring = str
+    xrange = range
+    StringIO = BytesIO
 
 AUDIO_FILE_EXT_ALIASES = {
     "m4a": "mp4"
@@ -17,6 +35,7 @@ AUDIO_FILE_EXT_ALIASES = {
 
 
 class AudioSegment(object):
+
     """
     Note: AudioSegment objects are immutable
 
@@ -36,7 +55,6 @@ class AudioSegment(object):
         else:
             # normal construction
             data = data if isinstance(data, basestring) else data.read()
-
             raw = wave.open(StringIO(data), 'rb')
 
             raw.rewind()
@@ -57,8 +75,13 @@ class AudioSegment(object):
         return round(1000 * (self.frame_count() / self.frame_rate))
 
     def __eq__(self, other):
-        try: return self._data == other._data
-        except: return False
+        try:
+            return self._data == other._data
+        except:
+            return False
+
+    def __ne__(self, other):
+        return not (self == other)
 
     def __iter__(self):
         return (self[i] for i in xrange(len(self)))
@@ -81,14 +104,14 @@ class AudioSegment(object):
 
         # ensure the output is as long as the requester is expecting
         expected_length = end - start
-        missing_frames = (expected_length - len(data)) / self.frame_width
+        missing_frames = (expected_length - len(data)) // self.frame_width
         if missing_frames:
             if missing_frames > self.frame_count(ms=2):
-                raise TooManyMissingFrames("You should never be filling in "\
-                "   more than 2 ms with silence here, missing frames: %s" % \
-                missing_frames)
+                raise TooManyMissingFrames("You should never be filling in "
+                                           "   more than 2 ms with silence here, missing frames: %s" %
+                                           missing_frames)
             silence = audioop.mul(data[:self.frame_width],
-                self.sample_width, 0)
+                                  self.sample_width, 0)
             data += (silence * missing_frames)
 
         return self._spawn(data)
@@ -101,8 +124,8 @@ class AudioSegment(object):
 
     def __sub__(self, arg):
         if isinstance(arg, AudioSegment):
-            raise TypeError("AudioSegment objects can't be subtracted from "\
-                "each other")
+            raise TypeError("AudioSegment objects can't be subtracted from "
+                            "each other")
         else:
             return self.apply_gain(-arg)
 
@@ -132,7 +155,7 @@ class AudioSegment(object):
         """
         # accept lists of data chunks
         if isinstance(data, list):
-            data = ''.join(data)
+            data = b''.join(data)
 
         # accept file-like objects
         if hasattr(data, 'read'):
@@ -141,11 +164,11 @@ class AudioSegment(object):
             data = data.read()
 
         metadata = {
-                    'sample_width': self.sample_width,
-                    'frame_rate': self.frame_rate,
-                    'frame_width': self.frame_width,
-                    'channels': self.channels
-                    }
+            'sample_width': self.sample_width,
+            'frame_rate': self.frame_rate,
+            'frame_width': self.frame_width,
+            'channels': self.channels
+        }
         metadata.update(overrides)
         return AudioSegment(data=data, metadata=metadata)
 
@@ -173,52 +196,52 @@ class AudioSegment(object):
             self.frame_count(ms=val)
         return int(val)
 
+    @classmethod
+    def empty(cls):
+        return cls(b'', metadata={"channels": 1, "sample_width": 1, "frame_rate": 1, "frame_width": 1})
+
     @classmethod
     def from_file(cls, file, format=None):
         file = _fd_or_path_or_tempfile(file, 'rb', tempfile=False)
 
-        if not format:
-            format = os.path.splitext(file.name)[1]
-            format = format.strip(".")
-            
-        format = AUDIO_FILE_EXT_ALIASES.get(format, format)
+        if format:
+            format = AUDIO_FILE_EXT_ALIASES.get(format, format)
 
         if format == 'wav':
             return cls.from_wav(file)
 
-        input = NamedTemporaryFile(mode='wb', delete=False)
-        input.write(file.read())
-        input.flush()
+        input_file = NamedTemporaryFile(mode='wb', delete=False)
+        input_file.write(file.read())
+        input_file.flush()
 
         output = NamedTemporaryFile(mode="rb", delete=False)
 
         ffmpeg_call = [cls.ffmpeg,
                        '-y',  # always overwrite existing files
                        ]
+
+        # If format is not defined
+        # ffmpeg will detect it automatically
         if format:
             ffmpeg_call += ["-f", format]
+
         ffmpeg_call += [
-                        "-i", input.name,  # input options (filename last)
-                        "-vn",  # Drop any video streams if there are any
+            "-i", input_file.name,  # input_file options (filename last)
+            "-vn",  # Drop any video streams if there are any
+            "-f", "wav",  # output options (filename last)
+            output.name
+        ]
 
-                        "-f", "wav",  # output options (filename last)
-                        output.name
-                        ]
+        subprocess.call(ffmpeg_call, stderr=open(os.devnull))
 
-        subprocess.call(ffmpeg_call,
-                        stderr=open(os.devnull)
-                        )
-        
         obj = cls.from_wav(output)
-        
-        input.close()
+
+        input_file.close()
         output.close()
-        os.unlink(input.name)
+        os.unlink(input_file.name)
         os.unlink(output.name)
-        
+
         return obj
-        
-        
 
     @classmethod
     def from_mp3(cls, file):
@@ -238,9 +261,36 @@ class AudioSegment(object):
         file.seek(0)
         return cls(data=file)
 
-    def export(self, out_f=None, format='mp3', codec=None):
+    def export(self, out_f=None, format='mp3', codec=None, bitrate=None, parameters=None, tags=None, id3v2_version='4'):
+        """
+        Export an AudioSegment to a file with given options
+
+        out_f (string):
+            Path to destination audio file
+
+        format (string)
+            Format for destination audio file. ('mp3', 'wav', 'ogg' or other ffmpeg supported files)
+
+        codec (string)
+            Codec used to encoding for the destination.
+
+        bitrate (string)
+            Bitrate used when encoding destination file. (128, 256, 312k...)
+
+        parameters (string)
+            Aditional ffmpeg parameters
+
+        tags (dict)
+            Set metadata information to destination files usually used as tags. ({title='Song Title', artist='Song Artist'})
+
+        id3v2_version (string)
+            Set ID3v2 version for tags. (default: '4')
+        """
+        id3v2_allowed_versions = ['3', '4']
+
         out_f = _fd_or_path_or_tempfile(out_f, 'wb+')
         out_f.seek(0)
+
         # for wav output we can just write the data directly to out_f
         if format == "wav":
             data = out_f
@@ -251,7 +301,9 @@ class AudioSegment(object):
         wave_data.setnchannels(self.channels)
         wave_data.setsampwidth(self.sample_width)
         wave_data.setframerate(self.frame_rate)
-        wave_data.setnframes(self.frame_count())
+        # For some reason packing the wave header struct with a float in python 2
+        # doesn't throw an exception
+        wave_data.setnframes(int(self.frame_count()))
         wave_data.writeframesraw(self._data)
         wave_data.close()
 
@@ -261,32 +313,61 @@ class AudioSegment(object):
 
         output = NamedTemporaryFile(mode="w+b", delete=False)
 
-        # build call args
-        args =[self.ffmpeg,
-            '-y',  # always overwrite existing files
-            "-f", "wav", "-i", data.name,  # input options (filename last)
-        ]
+        # build ffmpeg command to export
+        ffmpeg_call = [self.ffmpeg,
+                       '-y',  # always overwrite existing files
+                       "-f", "wav", "-i", data.name,  # input options (filename last)
+                       ]
+
+        if format == "ogg" and codec is None:
+            ffmpeg_call.extend(["-acodec", "libvorbis"])
+
         if codec is not None:
             # force audio encoder
-            args.extend(["-acodec", codec])
-        args.extend([
+            ffmpeg_call.extend(["-acodec", codec])
+
+        if bitrate is not None:
+            ffmpeg_call.extend(["-b:a", bitrate])
+
+        if parameters is not None:
+            # extend arguments with arbitrary set
+            ffmpeg_call.extend(parameters)
+
+        if tags is not None:
+            if not isinstance(tags, dict):
+                raise InvalidTag("Tags must be a dictionary.")
+            else:
+                # Extend ffmpeg command with tags
+                [ffmpeg_call.extend(['-metadata', '{0}={1}'.format(k, v)])
+                 for k, v in tags.items()]
+
+                if format == 'mp3':
+                    if id3v2_version not in id3v2_allowed_versions:
+                        raise InvalidID3TagVersion(
+                            "id3v2_version not allowed, allowed versions: %s" % id3v2_allowed_versions)
+                    ffmpeg_call.extend([
+                        "-id3v2_version",  id3v2_version
+                    ])
+
+        ffmpeg_call.extend([
             "-f", format, output.name,  # output options (filename last)
         ])
+
         # read stdin / write stdout
-        subprocess.call(args,
-            # make ffmpeg shut up
-            stderr=open(os.devnull)
-        )
+        subprocess.call(ffmpeg_call,
+                        # make ffmpeg shut up
+                        stderr=open(os.devnull)
+                        )
 
         output.seek(0)
         out_f.write(output.read())
-        
+
         data.close()
         output.close()
-        
+
         os.unlink(data.name)
         os.unlink(output.name)
-        
+
         out_f.seek(0)
         return out_f
 
@@ -303,16 +384,21 @@ class AudioSegment(object):
         if ms is not None:
             return ms * (self.frame_rate / 1000.0)
         else:
-            return float(len(self._data) / self.frame_width)
+            return float(len(self._data) // self.frame_width)
 
     def set_frame_rate(self, frame_rate):
         if frame_rate == self.frame_rate:
             return self
 
-        converted, _ = audioop.ratecv(self._data, self.sample_width,
-            self.channels, self.frame_rate, frame_rate, None)
+        if self._data:
+            converted, _ = audioop.ratecv(self._data, self.sample_width,
+                                          self.channels, self.frame_rate,
+                                          frame_rate, None)
+        else:
+            converted = self._data
+
         return self._spawn(data=converted,
-            overrides={'frame_rate': frame_rate})
+                           overrides={'frame_rate': frame_rate})
 
     def set_channels(self, channels):
         if channels == self.channels:
@@ -323,43 +409,37 @@ class AudioSegment(object):
             frame_width = self.frame_width * 2
         elif channels == 1 and self.channels == 2:
             fn = 'tomono'
-            frame_width = self.frame_width / 2
+            frame_width = self.frame_width // 2
 
         fn = getattr(audioop, fn)
         converted = fn(self._data, self.sample_width, 1, 1)
 
         return self._spawn(data=converted, overrides={'channels': channels,
-            'frame_width': frame_width})
+                                                      'frame_width': frame_width})
 
     @property
     def rms(self):
         return audioop.rms(self._data, self.sample_width)
-    
+
     @property
     def max(self):
         return audioop.max(self._data, self.sample_width)
-    
+
     @property
     def max_possible_amplitude(self):
-        bits = self.sample_width*8
-        max_possible_val = (2**bits)
-        
+        bits = self.sample_width * 8
+        max_possible_val = (2 ** bits)
+
         # since half is above 0 and half is below the max amplitude is divided
         return max_possible_val / 2
-    
-    def normalize(self, headroom=0.1):
-        """
-        headroom is how close to the maximum volume to boost the signal up to (specified in dB)
-        """
-        peak_sample_val = self.max
-        target_peak = self.max_possible_amplitude * db_to_float(-headroom)
-        
-        needed_boost = ratio_to_db(target_peak / peak_sample_val)
-        return self.apply_gain(needed_boost)
+
+    @property
+    def duration_seconds(self):
+        return self.frame_rate and self.frame_count() / self.frame_rate or 0.0
 
     def apply_gain(self, volume_change):
         return self._spawn(data=audioop.mul(self._data, self.sample_width,
-            db_to_float(float(volume_change))))
+                                            db_to_float(float(volume_change))))
 
     def overlay(self, seg, position=0, loop=False):
         output = TemporaryFile()
@@ -384,7 +464,7 @@ class AudioSegment(object):
                 loop = False
 
             output.write(audioop.add(seg1[pos:pos + seg2_len], seg2,
-                sample_width))
+                                     sample_width))
             pos += seg2_len
 
             if not loop:
@@ -413,7 +493,7 @@ class AudioSegment(object):
         return seg1._spawn(data=output)
 
     def fade(self, to_gain=0, from_gain=0, start=None, end=None,
-            duration=None):
+             duration=None):
         """
         Fade the volume of this audio segment.
 
@@ -433,8 +513,8 @@ class AudioSegment(object):
             the duration of the fade
         """
         if None not in [duration, end, start]:
-            raise TypeError('Only two of the three arguments, "start", '\
-                '"end", and "duration" may be specified')
+            raise TypeError('Only two of the three arguments, "start", '
+                            '"end", and "duration" may be specified')
 
         # no fade == the same audio
         if to_gain == 0 and from_gain == 0:
@@ -442,8 +522,8 @@ class AudioSegment(object):
 
         frames = self.frame_count()
 
-        start = min(len(self), start)
-        end = min(len(self), end)
+        start = min(len(self), start) if start is not None else None
+        end = min(len(self), end) if end is not None else None
 
         if start is not None and start < 0:
             start += len(self)
@@ -469,24 +549,41 @@ class AudioSegment(object):
         before_fade = self[:start]._data
         if from_gain != 0:
             before_fade = audioop.mul(before_fade, self.sample_width,
-                from_power)
+                                      from_power)
         output.append(before_fade)
 
         gain_delta = db_to_float(to_gain) - from_power
-        scale_step = gain_delta / duration
 
-        for i in range(duration):
-            volume_change = from_power + (scale_step * i)
-            chunk = self[start + i]
-            chunk = audioop.mul(chunk._data, self.sample_width, volume_change)
+        # fades longer than 100ms can use coarse fading (one gain step per ms),
+        # shorter fades will have audible clicks so they use precise fading (one
+        # gain step per sample)
+        if duration > 100:
+            scale_step = gain_delta / duration
 
-            output.append(chunk)
+            for i in range(duration):
+                volume_change = from_power + (scale_step * i)
+                chunk = self[start + i]
+                chunk = audioop.mul(chunk._data, self.sample_width, volume_change)
+
+                output.append(chunk)
+        else:
+            start_frame = self.frame_count(ms=start)
+            end_frame = self.frame_count(ms=end)
+            fade_frames = end_frame - start_frame
+            scale_step = gain_delta / fade_frames
+
+            for i in range(int(fade_frames)):
+                volume_change = from_power + (scale_step * i)
+                sample = self.get_frame(int(start_frame + i))
+                sample = audioop.mul(sample, self.sample_width, volume_change)
+
+                output.append(sample)
 
         # original data after the crossfade portion, at the new volume
         after_fade = self[end:]._data
         if to_gain != 0:
             after_fade = audioop.mul(after_fade, self.sample_width,
-                db_to_float(to_gain))
+                                     db_to_float(to_gain))
         output.append(after_fade)
 
         return self._spawn(data=output)
@@ -496,10 +593,11 @@ class AudioSegment(object):
 
     def fade_in(self, duration):
         return self.fade(from_gain=-120, duration=duration, start=0)
-    
+
     def reverse(self):
         return self._spawn(
             data=audioop.reverse(self._data, self.sample_width)
         )
-        
-    
+
+
+from . import effects
